from sklearn.base import BaseEstimator, ClusterMixin
from sklearn.utils.validation import assert_all_finite, check_array, check_is_fitted
from sklearn.neighbors import DistanceMetric
from ..errors.errors import SearchFailedError, TreeNotFoundError
from sklearn.exceptions import NotFittedError
from ..predictors.predictor import DL85Predictor
import numpy as np


class DL85Cluster(DL85Predictor, BaseEstimator, ClusterMixin):
    """ An optimal binary decision tree classifier.

    Parameters
    ----------
    max_depth : int, default=1
        Maximum depth of the tree to be found
    min_sup : int, default=1
        Minimum number of examples per leaf
    iterative : bool, default=False
        Whether the search will be Iterative Deepening Search or not. By default, it is Depth First Search
    max_error : int, default=0
        Maximum allowed error. Default value stands for no bound. If no tree can be found that is strictly better, the model remains empty.
    stop_after_better : bool, default=False
        A parameter used to indicate if the search will stop after finding a tree better than max_error
    time_limit : int, default=0
        Allocated time in second(s) for the search. Default value stands for no limit. The best tree found within the time limit is stored, if this tree is better than max_error.
    verbose : bool, default=False
        A parameter used to switch on/off the print of what happens during the search
    desc : bool, default=False
        A parameter used to indicate if the sorting of the items is done in descending order of information gain
    asc : bool, default=False
        A parameter used to indicate if the sorting of the items is done in ascending order of information gain
    repeat_sort : bool, default=False
        A parameter used to indicate whether the sorting of items is done at each level of the lattice or only before the search
    nps : bool, default=False
        A parameter used to indicate if only optimal solutions should be stored in the cache.
    print_output : bool, default=False
        A parameter used to indicate if the search output will be printed or not

    Attributes
    ----------
    tree_ : str
        Outputted tree in serialized form; remains empty as long as no model is learned.
    size_ : int
        The size of the outputted tree
    depth_ : int
        Depth of the found tree
    error_ : float
        Error of the found tree
    accuracy_ : float
        Accuracy of the found tree on training set
    lattice_size_ : int
        The number of nodes explored before found the optimal tree
    runtime_ : float
        Time of the optimal decision tree search
    timeout_ : bool
        Whether the search reached timeout or not
    classes_ : ndarray, shape (n_classes,)
        The classes seen at :meth:`fit`.
    """

    def __init__(
            self,
            max_depth=1,
            min_sup=1,
            error_function=None,
            iterative=False,
            max_error=0,
            stop_after_better=False,
            time_limit=0,
            verbose=False,
            # desc_sort_function=None,
            # asc_sort_function=None,
            repeat_sort=False,
            leaf_value_function=None,
            nps=False,
            print_output=False):

        DL85Predictor.__init__(self,
                               max_depth,
                               min_sup,
                               error_function,
                               iterative,
                               max_error,
                               stop_after_better,
                               time_limit,
                               verbose,
                               None,
                               None,
                               repeat_sort,
                               leaf_value_function,
                               nps,
                               print_output)

    @staticmethod
    def default_error(tids, X):
        dist = DistanceMetric.get_metric('euclidean')
        X_subset = np.asarray([X[index, :] for index in list(tids)], dtype='int32')
        centroid = np.mean(X_subset, axis=0).reshape(1, X_subset.shape[1])
        distances = [dist.pairwise(instance.reshape(1, X_subset.shape[1]), centroid)[0, 0] for instance in X_subset]
        return sum(distances)

    @staticmethod
    def default_leaf_value(tids, X):
        return np.mean(X.take(list(tids)))

    def _more_tags(self):
        return {'X_types': 'categorical',
                'allow_nan': False}

    def fit(self, X, X_error=None):
        """Implements the standard fitting function for a DL8.5 classifier.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            The training input samples. If X_error is provided, it represents explanation input
        X_error : array-like, shape (n_samples, n_features_1)
            The training input used to calculate error. If it is not provided X is used to calculate error

        Returns
        -------
        self : object
            Returns self.
        """

        # Check that X has correct shape and raise ValueError if not
        assert_all_finite(X)
        X = check_array(X, dtype='int32')

        # Check that X_error has correct shape and raise ValueError if not
        if X_error is not None:
            assert_all_finite(X_error)
            X_error = check_array(X_error, dtype='int32')

        if self.error_function is None:
            if X_error is None:
                self.error_function = lambda tids: self.default_error(tids, X)
            else:
                if X_error.shape[0] == X.shape[0]:
                    self.error_function = lambda tids: self.default_error(tids, X_error)
                else:
                    raise ValueError("X_error does not have the same number of rows as X")

        if self.leaf_value_function is None:
            if X_error is None:
                self.leaf_value_function = lambda tids: self.default_leaf_value(tids, X)
            else:
                if X_error.shape[0] == X.shape[0]:
                    self.leaf_value_function = lambda tids: self.default_leaf_value(tids, X_error)
                else:
                    raise ValueError("X_error does not have the same number of rows as X")

        # call fit method of the predictor
        self.predictor_fit(X)
        #print(self.tree_)

        # Return the classifier
        return self

    def predict(self, X):
        """ Implements the standard predict function for a DL8.5 classifier.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            The input samples.

        Returns
        -------
        y : ndarray, shape (n_samples,)
            The label for each sample is the label of the closest sample
            seen during fit.
        """

        # Check is fit had been called
        # check_is_fitted(self, attributes='tree_') # use of attributes is deprecated. alternative solution is below

        if hasattr(self, 'sol_size') is False:  # actually this case is not possible.
            raise NotFittedError("Call fit method first" % {'name': type(self).__name__})

        if self.tree_ is None:
            raise TreeNotFoundError("predict(): ", "Tree not found during training by DL8.5")

        if hasattr(self, 'tree_') is False:  # normally this case is not possible.
            raise SearchFailedError("PredictionError: ", "DL8.5 training has failed. Please contact the developers if the problem is in the scope claimed by the tool.")

        # Input validation
        X = check_array(X)

        self.y_ = self.predictor_predict(X)

        return self.y_





from .supervised.classifiers.classifier import DL85Classifier
from .predictors.predictor import DL85Predictor
from .unsupervised.clustering import DL85Cluster
from ._version import __version__

__all__ = ['__version__']



from libcpp cimport bool
from libcpp.vector cimport vector
from libcpp.stack cimport stack
from cython.operator cimport dereference as deref, preincrement as inc

cdef extern from "dataManager.h":
    cdef cppclass DataManager:
        int getNClasses()
        int* getSupports()

cdef extern from "rCover.h":
    cdef cppclass RCover:
        cppclass iterator:
            int wordIndex
            int operator*()
            iterator operator++()
            bool operator==(iterator)
            bool operator!=(iterator)
        iterator begin(bool trans_loop)
        iterator end(bool trans_loop)
        DataManager* dm
        stack[int] limit

cdef class ArrayIterator:
    cdef RCover* arr
    cdef RCover.iterator it
    cdef bool trans_loop

    def __init__(self, trans):
        self.trans_loop = trans

    def __iter__(self):
        return self

    def __next__(self):
        if self.trans_loop:
            if self.it.wordIndex < self.arr.limit.top():
                val = deref(self.it)
                self.it = inc(self.it)
                return val
            else:
                raise StopIteration()
        else:
            if self.it.wordIndex < deref(self.arr.dm).getNClasses():
                val = deref(self.it)
                self.it = inc(self.it)
                return val
            else:
                raise StopIteration()

    def init_iterator(self):
            self.it = self.arr.begin(self.trans_loop)

cdef public wrap_array(RCover *ar, bool trans):
    tid_python_object = ArrayIterator(trans)
    tid_python_object.arr = ar
    tid_python_object.init_iterator()
    return tid_python_object

cdef public vector[float] call_python_error_function(python_function, RCover *ar):
    return python_function(wrap_array(ar, True))

cdef public vector[float] call_python_fast_error_function(python_fast_function, RCover *ar):
    return python_fast_function(wrap_array(ar, False))

cdef public float call_python_predictor_error_function(python_predictor_function, RCover *ar):
    return python_predictor_function(wrap_array(ar, True))




from sklearn.base import BaseEstimator, ClassifierMixin
from sklearn.utils.validation import check_X_y, check_array, check_is_fitted
from sklearn.utils.multiclass import unique_labels
from sklearn.exceptions import NotFittedError
from ...errors.errors import SearchFailedError, TreeNotFoundError
import json


class DL85Classifier(BaseEstimator, ClassifierMixin):
    """ An optimal binary decision tree classifier.

    Parameters
    ----------
    max_depth : int, default=1
        Maximum depth of the tree to be found
    min_sup : int, default=1
        Minimum number of examples per leaf
    iterative : bool, default=False
        Whether the search will be Iterative Deepening Search or not. By default, it is Depth First Search
    max_error : int, default=0
        Maximum allowed error. Default value stands for no bound. If no tree can be found that is strictly better, the model remains empty.
    stop_after_better : bool, default=False
        A parameter used to indicate if the search will stop after finding a tree better than max_error
    time_limit : int, default=0
        Allocated time in second(s) for the search. Default value stands for no limit. The best tree found within the time limit is stored, if this tree is better than max_error.
    verbose : bool, default=False
        A parameter used to switch on/off the print of what happens during the search
    desc : bool, default=False
        A parameter used to indicate if the sorting of the items is done in descending order of information gain
    asc : bool, default=False
        A parameter used to indicate if the sorting of the items is done in ascending order of information gain
    repeat_sort : bool, default=False
        A parameter used to indicate whether the sorting of items is done at each level of the lattice or only before the search
    nps : bool, default=False
        A parameter used to indicate if only optimal solutions should be stored in the cache.
    print_output : bool, default=False
        A parameter used to indicate if the search output will be printed or not

    Attributes
    ----------
    tree_ : str
        Outputted tree in serialized form; remains empty as long as no model is learned.
    size_ : int
        The size of the outputted tree
    depth_ : int
        Depth of the found tree
    error_ : float
        Error of the found tree
    accuracy_ : float
        Accuracy of the found tree on training set
    lattice_size_ : int
        The number of nodes explored before found the optimal tree
    runtime_ : float
        Time of the optimal decision tree search
    timeout_ : bool
        Whether the search reached timeout or not
    classes_ : ndarray, shape (n_classes,)
        The classes seen at :meth:`fit`.
    """

    def __init__(
            self,
            max_depth=1,
            min_sup=1,
            error_function=None,
            fast_error_function=None,
            iterative=False,
            max_error=0,
            stop_after_better=False,
            time_limit=0,
            verbose=False,
            desc=False,
            asc=False,
            repeat_sort=False,
            nps=False,
            print_output=False):
        self.error_function = error_function
        self.fast_error_function = fast_error_function
        self.max_depth = max_depth
        self.min_sup = min_sup
        self.iterative = iterative
        self.max_error = max_error
        self.stop_after_better = stop_after_better
        self.time_limit = time_limit
        self.verbose = verbose
        self.desc = desc
        self.asc = asc
        self.repeat_sort = repeat_sort
        self.nps = nps
        self.print_output = print_output

    def _more_tags(self):
        return {'X_types': 'categorical',
                'allow_nan': False}

    def fit(self, X, y):
        """Implements the standard fitting function for a DL8.5 classifier.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            The training input samples.
        y : array-like, shape (n_samples,)
            The target values. An array of int.

        Returns
        -------
        self : object
            Returns self.
        """

        # Check that X and y have correct shape and raise ValueError if not
        X, y = check_X_y(X, y, dtype='int32')

        # sys.path.insert(0, "../../")
        import dl85Optimizer
        solution = dl85Optimizer.solve(data=X,
                                       target=y,
                                       func=self.error_function,
                                       fast_func=self.fast_error_function,
                                       predictor_func=None,
                                       max_depth=self.max_depth,
                                       min_sup=self.min_sup,
                                       max_error=self.max_error,
                                       stop_after_better=self.stop_after_better,
                                       iterative=self.iterative,
                                       time_limit=self.time_limit,
                                       verb=self.verbose,
                                       desc=self.desc,
                                       asc=self.asc,
                                       repeat_sort=self.repeat_sort,
                                       bin_save=False,
                                       nps=self.nps)

        # print(solution)
        solution = solution.splitlines()
        self.sol_size = len(solution)

        # if self.sol_size_ == 1:
        #     raise ValueError(solution[0])

        if self.sol_size == 8 or self.sol_size == 9:  # solution found
            self.tree_ = json.loads(solution[1].split('Tree: ')[1])
            self.size_ = int(solution[2].split(" ")[1])
            self.depth_ = int(solution[3].split(" ")[1])
            self.error_ = float(solution[4].split(" ")[1])
            if self.size_ < 3 and self.max_error > 0:
                self.accuracy_ = -1
            else:
                self.accuracy_ = float(solution[5].split(" ")[1])

            if self.sol_size == 8:  # without timeout
                if self.size_ < 3 and self.max_error > 0:
                    print("DL8.5 fitting: Solution found but not kept since it has the same error as the max error you specify as unreachable.")
                    self.tree_ = None
                    self.size_ = -1
                    self.depth_ = -1
                    self.error_ = -1
                    self.accuracy_ = -1
                else:
                    print("DL8.5 fitting: Solution found")

                self.lattice_size_ = int(solution[6].split(" ")[1])
                self.runtime_ = float(solution[7].split(" ")[1])
                self.timeout_ = False
            else:  # timeout reached
                if self.size_ < 3 and self.max_error > 0:
                    print("DL8.5 fitting: Timeout reached and solution found is not kept since it has the same error as the max error you specify as unreachable.")
                    self.tree_ = None
                    self.size_ = -1
                    self.depth_ = -1
                    self.error_ = -1
                    self.accuracy_ = -1
                else:
                    print("DL8.5 fitting: Timeout reached but solution found")

                self.lattice_size_ = int(solution[7].split(" ")[1])
                self.runtime_ = float(solution[8].split(" ")[1])
                self.timeout_ = True

            # Store the classes seen during fit
            self.classes_ = unique_labels(y)

        elif self.sol_size == 4 or self.sol_size == 5:  # solution not found
            self.tree_ = None
            self.size_ = -1
            self.depth_ = -1
            self.error_ = -1
            self.accuracy_ = -1
            if self.sol_size == 4:  # without timeout
                print("DL8.5 fitting: Solution not found")
                self.lattice_size_ = int(solution[2].split(" ")[1])
                self.runtime_ = float(solution[3].split(" ")[1])
                self.timeout_ = False
            else:  # timeout reached
                print("DL8.5 fitting: Timeout reached and solution not found")
                self.lattice_size_ = int(solution[3].split(" ")[1])
                self.runtime_ = float(solution[4].split(" ")[1])
                self.timeout_ = True

        if self.print_output:
            print(solution[0])
            print("Tree:", self.tree_)
            print("Size:", str(self.size_))
            print("Depth:", str(self.depth_))
            print("Error:", str(self.error_))
            print("Accuracy:", str(self.accuracy_))
            print("LatticeSize:", str(self.lattice_size_))
            print("Runtime:", str(self.runtime_))
            print("Timeout:", str(self.timeout_))

        # Return the classifier
        return self

    def predict(self, X):
        """ Implements the standard predict function for a DL8.5 classifier.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            The input samples.

        Returns
        -------
        y : ndarray, shape (n_samples,)
            The label for each sample is the label of the closest sample
            seen during fit.
        """

        # Check is fit had been called
        # check_is_fitted(self, attributes='tree_') # use of attributes is deprecated. alternative solution is below

        if hasattr(self, 'sol_size') is False:  # actually this case is not possible.
            raise NotFittedError("Call fit method first" % {'name': type(self).__name__})

        if self.tree_ is None:
            raise TreeNotFoundError("predict(): ", "Tree not found during training by DL8.5 — Check fitting message for more info.")

        if hasattr(self, 'tree_') is False:  # normally this case is not possible.
            raise SearchFailedError("PredictionError: ", "DL8.5 training has failed. Please contact the developers if the problem is in the scope claimed by the tool.")

        # Input validation
        X = check_array(X)

        self.y_ = []

        for i in range(X.shape[0]):
            self.y_.append(self.pred_on_dict(X[i, :]))

        return self.y_

    def pred_on_dict(self, instance):
        node = self.tree_
        while self.is_leaf_node(node) is not True:
            if instance[node['feat']] == 1:
                node = node['left']
            else:
                node = node['right']
        return node['class']

    @staticmethod
    def is_leaf_node(node):
        names = [x[0] for x in node.items()]
        return 'error' in names



from sklearn.utils.estimator_checks import check_estimator
from sklearn.model_selection import train_test_split
from ..classifier import DL85Classifier
import numpy as np
from random import randrange
from os import listdir
from os.path import isfile, join
from sklearn.metrics import accuracy_score


def test_fit():
    dataset = np.genfromtxt("datasets/anneal.txt", delimiter=' ')
    X = dataset[:, 1:]
    y = dataset[:, 0]
    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.2, random_state=0)
    clf1 = DL85Classifier(max_depth=randrange(1, 4), min_sup=randrange(1, X_train.shape[0] // 4))
    clf1.fit(X_train, y_train)

    assert clf1.sol_size in [4, 5, 8, 9]


def test_predict():
    dataset = np.genfromtxt("datasets/anneal.txt", delimiter=' ')
    X = dataset[:, 1:]
    y = dataset[:, 0]
    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.2, random_state=0)
    clf1 = DL85Classifier(max_depth=randrange(1, 4), min_sup=randrange(1, X_train.shape[0] // 4))
    clf1.fit(X_train, y_train)

    if clf1.sol_size in [8, 9]:
        y_pred1 = clf1.predict(X_test)

        def is_class(y_pred):
            for i in y_pred:
                if i not in list(set(list(clf1.classes_))):
                    return False
            return True

        assert len(y_pred1) == X_test.shape[0] and is_class(
            y_pred1) is True  # list(set(y_pred1)) == list(set(list(clf1.classes_)))
    else:
        assert clf1.sol_size in [4, 5]


def test_depth_2():
    solutions = [137, 10, 87, 22, 177, 267, 60, 16, 70, 32, 418, 599, 22, 252, 153, 58, 9, 55, 508, 282, 75, 17, 437, 0]
    mypath = "./datasets"
    onlyfiles = [f for f in listdir(mypath) if isfile(join(mypath, f))]
    onlyfiles = sorted(onlyfiles)
    depth = 2

    for i, file in enumerate(onlyfiles):
        dataset = np.genfromtxt(join(mypath, file), delimiter=' ')
        X = dataset[:, 1:]
        y = dataset[:, 0]
        X = X.astype('int32')
        y = y.astype('int32')

        clf = DL85Classifier(max_depth=depth, time_limit=600)
        clf.fit(X, y)
        y_pred = clf.predict(X)
        assert clf.depth_ <= depth
        assert clf.size_ <= 2 ** (depth+1) - 1
        assert clf.error_ <= solutions[i]
        assert clf.error_ == int(X.shape[0] - X.shape[0] * accuracy_score(y, y_pred))


def test_depth_3():
    solutions = [112, 5, 73, 15, 162, 236, 41, 10, 61, 22, 198, 369, 12, 8, 47, 46, 0, 29, 224, 216, 26, 12, 403, 0]
    mypath = "./datasets"
    onlyfiles = [f for f in listdir(mypath) if isfile(join(mypath, f))]
    onlyfiles = sorted(onlyfiles)
    depth = 3

    for i, file in enumerate(onlyfiles):
        if file in ["ionosphere.txt", "letter.txt", "mushroom.txt", "pendigits.txt", "splice-1.txt", "vehicle.txt"]:
            continue
        dataset = np.genfromtxt(join(mypath, file), delimiter=' ')
        X = dataset[:, 1:]
        y = dataset[:, 0]
        X = X.astype('int32')
        y = y.astype('int32')

        clf = DL85Classifier(max_depth=depth, time_limit=600)
        clf.fit(X, y)
        y_pred = clf.predict(X)
        assert clf.depth_ <= depth
        assert clf.size_ <= 2 ** (depth+1) - 1
        assert clf.error_ <= solutions[i]
        assert clf.error_ == int(X.shape[0] - X.shape[0] * accuracy_score(y, y_pred))


check_estimator(DL85Classifier)



from setuptools import find_packages, setup
from setuptools import Extension
from Cython.Build import cythonize
import platform
import codecs
from dl85 import __version__

DISTNAME = 'dl8.5'
DESCRIPTION = 'A package to build an optimal binary decision tree classifier.'
with codecs.open('README.rst', encoding='utf-8-sig') as f:
    LONG_DESCRIPTION = f.read()
AUTHORS = 'Gael Aglin, Siegfried Nijssen, Pierre Schaus'
AUTHORS_EMAIL = 'aglingael@gmail.com, siegfried.nijssen@gmail.com, pschaus@gmail.com'
URL = 'https://github.com/aglingael/dl8.5'
LICENSE = 'LICENSE.txt'
DOWNLOAD_URL = 'https://github.com/aglingael/dl8.5'
VERSION = __version__
INSTALL_REQUIRES = ['setuptools', 'cython', 'numpy', 'scikit-learn']
SETUP_REQUIRES = ['cython']
KEYWORDS = ['decision trees', 'discrete optimization', 'classification']
CLASSIFIERS = ['Programming Language :: Python :: 3',
               'License :: OSI Approved :: MIT License',
               'Operating System :: OS Independent',
               'Intended Audience :: Science/Research',
               'Intended Audience :: Developers',
               'Topic :: Software Development',
               'Topic :: Scientific/Engineering']
EXTRAS_REQUIRE = {
    'tests': [
        'pytest',
        'pytest-cov'],
    'docs': [
        'sphinx',
        'sphinx-gallery',
        'sphinx_rtd_theme',
        'numpydoc',
        'sphinxcontrib',
        'matplotlib'
    ]
}
PROJECT_URLS = {
    "Source on github": "https://github.com/aglingael/dl8.5",
    "Documentation": "https://dl85.readthedocs.io/en/latest/?badge=latest",
}

EXTENSION_NAME = 'dl85Optimizer'
EXTENSION_LANGUAGE = 'c++'
EXTENSION_SOURCE_FILES = ['wrapping/src/headers/error_function.pyx',
                          'wrapping/dl85Optimizer.pyx',
                'wrapping/src/codes/data.cpp',
                'wrapping/src/codes/dataContinuous.cpp',
                'wrapping/src/codes/dataBinary.cpp',
                'wrapping/src/codes/dataManager.cpp',
                'wrapping/src/codes/rCover.cpp',
                'wrapping/src/codes/dl85.cpp',
                'wrapping/src/codes/experror.cpp',
                'wrapping/src/codes/globals.cpp',
                'wrapping/src/codes/lcm_pruned.cpp',
                'wrapping/src/codes/lcm_iterative.cpp',
                'wrapping/src/codes/query.cpp',
                'wrapping/src/codes/query_best.cpp',
                'wrapping/src/codes/query_totalfreq.cpp',
                'wrapping/src/codes/trie.cpp',
                'wrapping/src/codes/dataBinaryPython.cpp']
EXTENSION_INCLUDE_DIR = ['wrapping/src/headers']
# EXTENSION_BUILD_ARGS = ['-std=c++11']
EXTENSION_BUILD_ARGS = ['-std=c++11', '-DCYTHON_PEP489_MULTI_PHASE_INIT=0']
if platform.system() == 'Darwin':
    EXTENSION_BUILD_ARGS.append('-mmacosx-version-min=10.12')

dl85_extension = Extension(
    name=EXTENSION_NAME,
    language=EXTENSION_LANGUAGE,
    sources=EXTENSION_SOURCE_FILES,
    include_dirs=EXTENSION_INCLUDE_DIR,  # path for headers
    extra_compile_args=EXTENSION_BUILD_ARGS,
    extra_link_args=EXTENSION_BUILD_ARGS
)

setup(
    name=DISTNAME,
    version=VERSION,
    url=URL,
    project_urls=PROJECT_URLS,
    author=AUTHORS,
    author_email=AUTHORS_EMAIL,
    maintainer=AUTHORS,
    maintainer_email=AUTHORS_EMAIL,
    download_url=DOWNLOAD_URL,
    license=LICENSE,
    packages=find_packages(),  # ["dl85", "dl85.classifiers", "dl85.errors"],
    keywords=KEYWORDS,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    classifiers=CLASSIFIERS,
    setup_requires=SETUP_REQUIRES,
    install_requires=INSTALL_REQUIRES,
    # extras_require=EXTRAS_REQUIRE,
    zip_safe=True,  # the package can run out of an .egg file
    ext_modules=cythonize(
        [dl85_extension],
        compiler_directives={"language_level": "3"}
    )
)



from libc.stdlib cimport malloc, free
from libcpp.string cimport string
from libcpp.map cimport map
from libcpp.utility cimport pair
from libcpp cimport bool
from libcpp.vector cimport vector
from libcpp.functional cimport function
import numpy as np

cdef extern from "src/headers/globals.h":
    cdef cppclass Array[T]:
        cppclass iterator:
            T operator*()
            iterator operator++()
            bool operator==(iterator)
            bool operator!=(iterator)
        iterator begin()
        iterator end()
        int getSize()


cdef extern from "src/headers/dl85.h":
    string search ( int* supports,
                    int ntransactions,
                    int nattributes,
                    int nclasses,
                    int *data,
                    int *target,
                    float maxError,
                    bool stopAfterError,
                    bool iterative,
                    PyErrorWrapper error_callback,
                    PyFastErrorWrapper fast_error_callback,
                    PyPredictorErrorWrapper predictor_error_callback,
                    bool error_is_null,
                    bool fast_error_is_null,
                    int maxdepth,
                    int minsup,
                    bool infoGain,
                    bool infoAsc,
                    bool repeatSort,
                    int timeLimit,
                    map[int, pair[int, int]]* continuousMap,
                    bool save,
                    bool nps_param,
                    bool verbose_param,
                    bool predict) except +

cdef extern from "src/headers/py_error_function_wrapper.h":
    cdef cppclass PyErrorWrapper:
        PyErrorWrapper()
        PyErrorWrapper(object) # define a constructor that takes a Python object
             # note - doesn't match c++ signature - that's fine!

cdef extern from "src/headers/py_fast_error_function_wrapper.h":
    cdef cppclass PyFastErrorWrapper:
        PyFastErrorWrapper()
        PyFastErrorWrapper(object) # define a constructor that takes a Python object
             # note - doesn't match c++ signature - that's fine!

cdef extern from "src/headers/py_predictor_error_function_wrapper.h":
    cdef cppclass PyPredictorErrorWrapper:
        PyPredictorErrorWrapper()
        PyPredictorErrorWrapper(object) # define a constructor that takes a Python object
             # note - doesn't match c++ signature - that's fine!


def solve(data,
          target,
          func=None,
          fast_func=None,
          predictor_func=None,
          max_depth=1,
          min_sup=1,
          max_error=0,
          stop_after_better=False,
          iterative=False,
          time_limit=0,
          verb=False,
          desc=False,
          asc=False,
          repeat_sort=False,
          continuousMap=None,
          bin_save=False,
          nps=False,
          predictor=False):

    cdef PyErrorWrapper f_user = PyErrorWrapper(func)
    error_null_flag = True
    if func is not None:
        error_null_flag = False

    cdef PyFastErrorWrapper f_user_fast = PyFastErrorWrapper(fast_func)
    fast_error_null_flag = True
    if fast_func is not None:
        fast_error_null_flag = False

    cdef PyPredictorErrorWrapper f_user_predictor = PyPredictorErrorWrapper(predictor_func)

    data = data.astype('int32')


    ntransactions, nattributes = data.shape
    classes, supports = np.unique(target, return_counts=True)
    nclasses = len(classes)
    supports = supports.astype('int32')


    data = data.transpose()
    # print(data)
    if np.array_equal(data, data.astype('bool')) is False:  # WARNING: maybe categorical (not binary) inputs will be supported in the future
        raise ValueError("Bad input type. DL8.5 actually only supports binary (0/1) inputs")
    if not data.flags['C_CONTIGUOUS']:
        data = np.ascontiguousarray(data) # Makes a contiguous copy of the numpy array.
    cdef int [:, ::1] data_view = data
    cdef int *data_matrix = &data_view[0][0]

    cdef int [::1] target_view
    cdef int *target_array = NULL
    if target is not None:
        target = target.astype('int32')
        if not target.flags['C_CONTIGUOUS']:
            target = np.ascontiguousarray(target) # Makes a contiguous copy of the numpy array.
        target_view = target
        target_array = &target_view[0]
    else:
        nclasses = 0


    if not supports.flags['C_CONTIGUOUS']:
        supports = np.ascontiguousarray(supports) # Makes a contiguous copy of the numpy array.
    cdef int [::1] supports_view = supports

    # max_err = max_error - 1  # because maxError but not be reached
    if max_error < 0:  # raise error when incompatibility between max_error value and stop_after_better value
        stop_after_better = False

    cont_map = NULL
    if continuousMap is not None:
        #cont_map must be defined properly
        cont_map = NULL

    info_gain = not (desc == False and asc == False)

    out = search(&supports_view[0],
                 ntransactions,
                 nattributes,
                 nclasses,
                 data_matrix,
                 target_array,
                 max_error,
                 stop_after_better,
                 iterative,
                 error_callback = f_user,
                 fast_error_callback = f_user_fast,
                 predictor_error_callback = f_user_predictor,
                 error_is_null = error_null_flag,
                 fast_error_is_null = fast_error_null_flag,
                 maxdepth = max_depth,
                 minsup = min_sup,
                 infoGain = info_gain,
                 infoAsc = asc,
                 repeatSort = repeat_sort,
                 timeLimit = time_limit,
                 continuousMap = NULL,
                 save = bin_save,
                 nps_param = nps,
                 verbose_param = verb,
                 predict = predictor)

    return out.decode("utf-8")